//This program was developed by me, Simon Rae.
// You are free to use this for personal use.
// It has been developed for use with the M5Stack Core 2 (https://docs.m5stack.com/en/core/Core2%20v1.1)
// which is a very flexible microcontroller-based development kit.

//This code should be uploaded to the device using the Arduino environment. If you know how to use this aready, then you should find the install process
// relatively painless - just make sure you ahve all the libraries listed below installed in the Arduino environment before you start. If you are unfamiliar
// with the Arduino environment, then I suggest you start here - https://www.arduino.cc/, and take your time!

#include <M5Core2.h>
#include <WiFi.h>
#include <HTTPClient.h>
#include <ArduinoJson.h>
#include <esp_sleep.h>
#include <time.h>

// Wi-Fi configuration
const char* ssid = "YOUR_SSID"; // Replace with your wifi networks SSID
const char* password = "wifi_password";  // Replace with your wifi password
const char* volumioIP = "Your_Volumio_IP"; // Replace with your Volumio IP address

// NTP Server for initial time sync
const char* ntpServer = "pool.ntp.org"; //if you are having difficulties getting a time for the RTC, you may want to replace this with a local NTP
const long gmtOffset_sec = 0;      // GMT offset in seconds (0 for GMT)
const int daylightOffset_sec = 3600; // Daylight savings offset (3600 seconds = 1 hour)

// Colors
#define BACKGROUND_COLOR 0x0000  // Black
#define TEXT_COLOR 0xFFFF        // White
#define BUTTON_COLOR 0xae1b      // Red
#define BUTTON_ACTIVE 0xF800     // Red
#define SLIDER_BG 0x7BEF         // Gray
#define SLIDER_FG 0xFFFf         // Green
#define BUTTON_TXT 0x0000        // Black
#define BATTERY_GREEN 0xae1b     // Green for good battery
#define BATTERY_YELLOW 0xFFE0    // Yellow for medium battery
#define BATTERY_RED 0xF800       // Red for low battery
#define CLOCK_COLOR 0xFFFF       // White for clock display

// Global variables
int currentVolume = 50;
bool isPlaying = false;
String currentTrack = "No track playing";
String currentArtist = "Unknown artist";
String currentAlbum = "Unknown album";

// RTC variables
RTC_TimeTypeDef RTCtime;
RTC_DateTypeDef RTCdate;
char timeString[9]; // HH:MM:SS + null terminator
bool timeInitialized = false;
unsigned long lastClockUpdateTime = 0;
const unsigned long clockUpdateInterval = 60000; // Update clock every second

// Battery monitoring variables
float batteryLevel = 0;
unsigned long lastBatteryCheckTime = 0;
const unsigned long batteryCheckInterval = 30000; // Check battery every 30 seconds

// Power-saving mode variables
unsigned long lastActivityTime = 0;
const unsigned long inactivityTimeout = 1 * 60 * 1000; // 5 minutes in milliseconds
bool isInPowerSavingMode = false;

// Backlight and touch interrupt pins
#define TOUCH_INT_PIN 39 // GPIO pin connected to the touch controller's interrupt pin

// Interface layout
#define INFO_X 10
#define INFO_Y 30
#define INFO_WIDTH (320 - INFO_X - 10)

// Clock layout
#define CLOCK_X 16
#define CLOCK_Y 2

// Button layout constants
#define BUTTON_WIDTH 100
#define BUTTON_HEIGHT 40
#define CONTROLS_Y 165
#define BUTTON_SPACING 10
#define BUTTON_START_X 0  // Starting X position for the first button
#define VOLUME_Y 225

// Battery indicator layout
#define BATTERY_X 280
#define BATTERY_Y 5
#define BATTERY_WIDTH 30
#define BATTERY_HEIGHT 15
#define BATTERY_TIP_WIDTH 4
#define BATTERY_TIP_HEIGHT 8

// Function to get the X position of a button by its index (0-3)
int getButtonX(int buttonIndex) {
  return BUTTON_START_X + buttonIndex * (BUTTON_WIDTH + BUTTON_SPACING);
}

// Touch handlers
bool checkButton(int x, int y, int btnX, int btnY, int btnW, int btnH) {
  return (x >= btnX && x <= btnX + btnW && y >= btnY && y <= btnY + btnH);
}

// Time functions
void initializeTime() {
  // Configure time using NTP
  configTime(gmtOffset_sec, daylightOffset_sec, ntpServer);
  
  // Get the current time
  struct tm timeinfo;
  if (getLocalTime(&timeinfo)) {
    // Set RTC using obtained time
    RTCtime.Hours = timeinfo.tm_hour;
    RTCtime.Minutes = timeinfo.tm_min;
    RTCtime.Seconds = timeinfo.tm_sec;
    M5.Rtc.SetTime(&RTCtime);
    
    RTCdate.Year = timeinfo.tm_year + 1900;
    RTCdate.Month = timeinfo.tm_mon + 1;
    RTCdate.Date = timeinfo.tm_mday;
    M5.Rtc.SetDate(&RTCdate);
    
    timeInitialized = true;
    Serial.println("Time synchronized with NTP server");
  } else {
    Serial.println("Failed to obtain time from NTP server");
  }
}

void updateTimeString() {
  M5.Rtc.GetTime(&RTCtime);
  sprintf(timeString, "%02d:%02d", RTCtime.Hours, RTCtime.Minutes);
}

// Volumio API functions
bool sendVolumeCommand(int volume) {
  if (WiFi.status() == WL_CONNECTED) {
    HTTPClient http;
    String url = "http://" + String(volumioIP) + "/api/v1/commands/?cmd=volume&volume=" + String(volume);
    http.begin(url);
    int httpResponseCode = http.GET();
    bool success = (httpResponseCode == 200);
    http.end();
    return success;
  }
  return false;
}

bool sendPlaybackCommand(String command) {
  if (WiFi.status() == WL_CONNECTED) {
    HTTPClient http;
    String url = "http://" + String(volumioIP) + "/api/v1/commands/?cmd=" + command;
    http.begin(url);
    int httpResponseCode = http.GET();
    bool success = (httpResponseCode == 200);
    http.end();
    return success;
  }
  return false;
}

bool getPlayerState() {
  if (WiFi.status() == WL_CONNECTED) {
    HTTPClient http;
    String url = "http://" + String(volumioIP) + "/api/v1/getState";
    http.begin(url);
    int httpResponseCode = http.GET();
    if (httpResponseCode == 200) {
      String payload = http.getString();
      DynamicJsonDocument doc(2048);
      DeserializationError error = deserializeJson(doc, payload);
      if (!error) {
        currentVolume = doc["volume"];
        String status = doc["status"];
        isPlaying = (status == "play");
        if (doc.containsKey("title")) {
          currentTrack = doc["title"].as<String>();
        } else {
          currentTrack = "No track playing";
        }
        if (doc.containsKey("artist")) {
          currentArtist = doc["artist"].as<String>();
        } else {
          currentArtist = "Unknown artist";
        }
        if (doc.containsKey("album")) {
          currentAlbum = doc["album"].as<String>();
        } else {
          currentAlbum = "Unknown album";
        }
        http.end();
        return true;
      }
    }
    http.end();
  }
  return false;
}

// Battery monitoring function
void updateBatteryLevel() {
  batteryLevel = M5.Axp.GetBatVoltage();
}

// Draw UI elements
void drawButton(int x, int y, int w, int h, String text, bool active = false) {
  M5.Lcd.fillRoundRect(x, y, w, h, 8, active ? BUTTON_ACTIVE : BUTTON_COLOR);
  M5.Lcd.setTextColor(BUTTON_TXT);
  M5.Lcd.setTextSize(2);
  int textWidth = text.length() * 12;
  if (textWidth > w - 10) {
    M5.Lcd.setTextSize(1);
    textWidth = text.length() * 6;
  }
  int textX = x + (w - textWidth) / 2;
  int textY = y + (h - 16) / 2;
  M5.Lcd.setCursor(textX, textY);
  M5.Lcd.print(text);
}

void drawVolumeSlider() {
  M5.Lcd.fillRoundRect(40, VOLUME_Y, 240, 10, 4, SLIDER_BG);
  int sliderPos = map(currentVolume, 10, 100, 0, 240);
  M5.Lcd.fillRoundRect(40, VOLUME_Y, sliderPos, 10, 4, SLIDER_FG);
}

void drawBatteryIndicator() {
  // Determine battery color based on level
  uint16_t batteryColor;
  int filledWidth;
  
  if (batteryLevel >= 3.9) {
    batteryColor = BATTERY_GREEN; // Good battery
    filledWidth = BATTERY_WIDTH * 0.9;
  } else if (batteryLevel >= 3.7) {
    batteryColor = BATTERY_YELLOW; // Medium battery
    filledWidth = BATTERY_WIDTH * 0.6;
  } else {
    batteryColor = BATTERY_RED; // Low battery
    filledWidth = BATTERY_WIDTH * 0.3;
  }
  
  // Draw battery outline
  M5.Lcd.drawRect(BATTERY_X, BATTERY_Y, BATTERY_WIDTH, BATTERY_HEIGHT, TEXT_COLOR);
  M5.Lcd.fillRect(BATTERY_X + BATTERY_WIDTH, BATTERY_Y + (BATTERY_HEIGHT - BATTERY_TIP_HEIGHT) / 2, 
                 BATTERY_TIP_WIDTH, BATTERY_TIP_HEIGHT, TEXT_COLOR);
  
  // Draw battery fill level
  M5.Lcd.fillRect(BATTERY_X + 2, BATTERY_Y + 2, filledWidth - 1, BATTERY_HEIGHT - 4, batteryColor);
  
  // Optional: Draw battery percentage text
  M5.Lcd.setTextSize(1);
  M5.Lcd.setTextColor(0x0000);
  M5.Lcd.setCursor(BATTERY_X+4 , BATTERY_Y + 4);
  int batteryPercentage = map(batteryLevel * 100, 340, 410, 0, 100); // Map voltage to percentage (3.4V-4.1V)
  batteryPercentage = constrain(batteryPercentage, 0, 100); // Keep within 0-100 range
  M5.Lcd.print(String(batteryPercentage) + "%");
}

void drawClock() {
  // Update the time string
  updateTimeString();

  // Calculate the width of the time string
  int textWidth = M5.Lcd.textWidth(timeString);

  // Clear the previous time display area
  M5.Lcd.fillRect(CLOCK_X, CLOCK_Y, textWidth-15, 16, BACKGROUND_COLOR); // Adjust height (16) as needed

  // Draw the new time
  M5.Lcd.setTextSize(1);
  M5.Lcd.setTextColor(CLOCK_COLOR);
  M5.Lcd.setCursor(CLOCK_X, CLOCK_Y + 4);
  M5.Lcd.print(timeString);
}

void drawTrackInfo() {
  M5.Lcd.fillRect(INFO_X, INFO_Y, INFO_WIDTH, 120, BACKGROUND_COLOR);
  String displayTrack = currentTrack;
  if (displayTrack.length() > 24) {
    displayTrack = displayTrack.substring(0, 22) + "...";
  }
  String displayArtist = currentArtist;
  if (displayArtist.length() > 24) {
    displayArtist = displayArtist.substring(0, 22) + "...";
  }
  String displayAlbum = currentAlbum;
  if (displayAlbum.length() > 24) {
    displayAlbum = displayAlbum.substring(0, 22) + "...";
  }
  M5.Lcd.setTextColor(TEXT_COLOR);
  M5.Lcd.setTextSize(1);
  M5.Lcd.setCursor(INFO_X, INFO_Y + 10);
  M5.Lcd.print("TRACK:");
  M5.Lcd.setTextSize(2);
  M5.Lcd.setCursor(INFO_X, INFO_Y + 25);
  M5.Lcd.print(displayTrack);
  M5.Lcd.setTextSize(1);
  M5.Lcd.setCursor(INFO_X, INFO_Y + 50);
  M5.Lcd.print("ARTIST:");
  M5.Lcd.setTextSize(2);
  M5.Lcd.setCursor(INFO_X, INFO_Y + 65);
  M5.Lcd.print(displayArtist);
  M5.Lcd.setTextSize(1);
  M5.Lcd.setCursor(INFO_X, INFO_Y + 90);
  M5.Lcd.print("ALBUM:");
  M5.Lcd.setTextSize(2);
  M5.Lcd.setCursor(INFO_X, INFO_Y + 105);
  M5.Lcd.print(displayAlbum);
}

void drawInterface() {
  M5.Lcd.fillScreen(BACKGROUND_COLOR);
  M5.Lcd.fillRoundRect(60, 0, 200, 20, 10, BUTTON_COLOR);
  M5.Lcd.setTextColor(BUTTON_TXT);
  M5.Lcd.setTextSize(2);
  M5.Lcd.setCursor(80, 3);
  M5.Lcd.print("VOLUMIO REMOTE");
  drawClock();
  drawBatteryIndicator();
  drawTrackInfo();
  drawButton(getButtonX(0), CONTROLS_Y, BUTTON_WIDTH, BUTTON_HEIGHT, "PREV");
  drawButton(getButtonX(1), CONTROLS_Y, BUTTON_WIDTH, BUTTON_HEIGHT, isPlaying ? "PAUSE" : "PLAY", isPlaying);
  drawButton(getButtonX(2), CONTROLS_Y, BUTTON_WIDTH, BUTTON_HEIGHT, "NEXT");
  drawVolumeSlider();
}

void updateButtonStates() {
  drawButton(getButtonX(1), CONTROLS_Y, BUTTON_WIDTH, BUTTON_HEIGHT, isPlaying ? "PAUSE" : "PLAY", isPlaying);
}

// Power-saving mode functions
void enterPowerSavingMode() {
  // Turn off the backlight
  M5.Axp.SetLcdVoltage(0);
  delay(100); // Small delay to ensure the backlight turns off completely

  // Turn off the display
  M5.Lcd.sleep(); // Put the display to sleep

  // Set the power-saving flag
  isInPowerSavingMode = true;

  // Configure touch interrupt for wake-up
  pinMode(TOUCH_INT_PIN, INPUT);
  esp_sleep_enable_ext0_wakeup((gpio_num_t)TOUCH_INT_PIN, LOW); // Wake up on LOW signal

  // Enter deep sleep
  esp_deep_sleep_start();
}

void wakeUpFromPowerSaving() {
  M5.Axp.SetLcdVoltage(2700);; // Turn on the backlight
  M5.Lcd.wakeup(); // Turn on the display
  isInPowerSavingMode = false;
  lastActivityTime = millis(); // Reset the inactivity timer
  drawInterface(); // Redraw the UI
}

void setup() {
  M5.begin();
  M5.Axp.SetLcdVoltage(2700);; // Ensure backlight is on initially
  Serial.begin(115200);
  Serial.print("Free heap: ");
  Serial.println(ESP.getFreeHeap());
  M5.Lcd.setRotation(1);
  M5.Lcd.fillScreen(BACKGROUND_COLOR);
  M5.Lcd.setTextColor(TEXT_COLOR);
  M5.Lcd.setTextSize(1);
  M5.Lcd.setCursor(20, 120);
  M5.Lcd.print("Connecting to Wi-Fi...");
  
  // Initial battery reading
  updateBatteryLevel();
  
  WiFi.begin(ssid, password);
  int attempt = 0;
  while (WiFi.status() != WL_CONNECTED && attempt < 20) {
    delay(500);
    Serial.print(".");
    attempt++;
  }
  if (WiFi.status() == WL_CONNECTED) {
    Serial.println("Wi-Fi connected");
    M5.Lcd.fillScreen(BACKGROUND_COLOR);
    M5.Lcd.setCursor(20, 120);
    M5.Lcd.print("Connected!");
    
    // Initialize time from NTP server
    initializeTime();
    
    if (getPlayerState()) {
      M5.Lcd.setCursor(20, 150);
      M5.Lcd.print("Player connected!");
    } else {
      M5.Lcd.setCursor(20, 150);
      M5.Lcd.print("Player not found!");
    }
    delay(1000);
  } else {
    Serial.println("Wi-Fi connection failed");
    M5.Lcd.fillScreen(BACKGROUND_COLOR);
    M5.Lcd.setCursor(20, 120);
    M5.Lcd.print("Wi-Fi connection failed!");
    delay(2000);
  }
  drawInterface();
}

void loop() {
  M5.update();
  if (M5.Touch.changed) {
    if (M5.Touch.points) {
      Point p = *(M5.Touch.point);
      if (isInPowerSavingMode) {
        wakeUpFromPowerSaving();
      } else {
        lastActivityTime = millis(); // Reset inactivity timer
        if (checkButton(p.x, p.y, getButtonX(0), CONTROLS_Y, BUTTON_WIDTH, BUTTON_HEIGHT)) {
          sendPlaybackCommand("prev");
        }
        if (checkButton(p.x, p.y, getButtonX(1), CONTROLS_Y, BUTTON_WIDTH, BUTTON_HEIGHT)) {
          if (isPlaying) {
            sendPlaybackCommand("pause");
          } else {
            sendPlaybackCommand("play");
          }
          isPlaying = !isPlaying;
          drawButton(getButtonX(1), CONTROLS_Y, BUTTON_WIDTH, BUTTON_HEIGHT, isPlaying ? "PAUSE" : "PLAY", isPlaying);
        }
        if (checkButton(p.x, p.y, getButtonX(2), CONTROLS_Y, BUTTON_WIDTH, BUTTON_HEIGHT)) {
          sendPlaybackCommand("next");
        }
        if (p.y >= VOLUME_Y && p.y <= VOLUME_Y + 30 && p.x >= 0 && p.x <= 280) {
          int newVolume = map(p.x, 0, 280, 0, 100);
          if (newVolume != currentVolume) {
            currentVolume = newVolume;
            drawVolumeSlider();
            sendVolumeCommand(currentVolume);
          }
        }
      }
    }
  }

  // Check for inactivity
  if (!isInPowerSavingMode && millis() - lastActivityTime > inactivityTimeout) {
    enterPowerSavingMode();
  }

  // Update clock display every second
  if (millis() - lastClockUpdateTime > clockUpdateInterval) {
    lastClockUpdateTime = millis();
    drawClock();
  }

  // Update battery level periodically
  if (millis() - lastBatteryCheckTime > batteryCheckInterval) {
    lastBatteryCheckTime = millis();
    float oldBatteryLevel = batteryLevel;
    updateBatteryLevel();
    
    // Only redraw if the battery level has changed significantly
    if (abs(oldBatteryLevel - batteryLevel) > 0.05) {
      drawBatteryIndicator();
    }
  }

  // Update player state periodically
  static unsigned long lastUpdateTime = 0;
  if (millis() - lastUpdateTime > 3000 && WiFi.status() == WL_CONNECTED) {
    lastUpdateTime = millis();
    bool wasPlaying = isPlaying;
    String prevTrack = currentTrack;
    String prevArtist = currentArtist;
    String prevAlbum = currentAlbum;
    int prevVolume = currentVolume;
    if (getPlayerState()) {
      if (wasPlaying != isPlaying) {
        drawButton(getButtonX(1), CONTROLS_Y, BUTTON_WIDTH, BUTTON_HEIGHT, isPlaying ? "PAUSE" : "PLAY", isPlaying);
      }
      if (prevTrack != currentTrack || prevArtist != currentArtist || prevAlbum != currentAlbum) {
        drawTrackInfo();
      }
      if (prevVolume != currentVolume) {
        drawVolumeSlider();
      }
    }
  }

  delay(50);
}